import os
import csv
import sqlite3


def get_connection():
    db_file = os.path.dirname(os.path.realpath(__file__)) + "/data/locs.db"
    conn = sqlite3.connect(db_file)
    conn.text_factory = lambda x: unicode(x, 'utf-8', 'ignore')

    return conn


# This is the module context so we don't need to establish a new connection 
# for every method call.
conn = get_connection()


def populate_db(conn):
    cur = conn.cursor()
    cur.execute("DROP TABLE IF EXISTS cities")    

    cur.execute("CREATE TABLE cities(geoname_id INTEGER, continent_code TEXT, continent TEXT, country_iso_code TEXT, country TEXT, region_iso_code TEXT, region TEXT, city TEXT, metro_code TEXT, time_zone TEXT)")
    cur.execute("CREATE INDEX index_region ON cities (region)")
    cur.execute("CREATE INDEX index_region_iso ON cities (region_iso_code)")
    cur.execute("CREATE INDEX index_country ON cities (country)")
    cur.execute("CREATE INDEX index_city ON cities (city)")
    cur_dir = os.path.dirname(os.path.realpath(__file__))
    with open(cur_dir+"/data/GeoLite2-City-Locations.csv", "rb") as info:
        reader = csv.reader(info)
        for row in reader:
            cur.execute("INSERT INTO cities VALUES(?, ?, ?, ?, ?, ?, ?, ?, ?, ?);", row)

        conn.commit()


def db_has_data(conn):
    cur = conn.cursor()

    cur.execute("SELECT Count(*) FROM sqlite_master WHERE name='cities';")
    data = cur.fetchone()[0]

    if data > 0:
        cur.execute("SELECT Count(*) FROM cities")
        data = cur.fetchone()[0]
        return data > 0

    return False


def get_places_by_type(place_name, place_type):
    place_name = str(place_name).strip().lower().title().replace('"','')
    
    try:
        conn = conn
    except:
        conn = get_connection()

    if not db_has_data(conn):
        populate_db(conn)

    cur = conn.cursor()
    cur.execute('SELECT geoname_id FROM cities WHERE ' + place_type + ' = "' + place_name + '" LIMIT 1')
    rows = cur.fetchall()

    if len(rows) > 0:
        return rows

    if len(place_name) == 2: #maybe it's an iso code
        cur.execute('SELECT geoname_id FROM cities WHERE ' + place_type + ' = "' + place_name.upper() + '" LIMIT 1')
        rows = cur.fetchall()

        if len(rows) > 0:
            return rows

    return []



from .value_checks import (is_a_date, is_a_int, is_a_bool, is_a_float, 
    is_a_coord, is_a_coord_pair, is_a_city)
from .list_checks import column_probability_for_type
import collections
from address import AddressParser

"""Takes a list of values and returns a type signature

:param row: a list of values, potentially of different types
:returns a list of types, like ['str', 'str', 'int', 'date']
"""
def row_simple_types(row):
    types = []
    for col in row:
        if is_a_date(col):
            types.append('date')
        elif is_a_float(col):
            types.append('float')
        else:
            try:
                int_col = int(col)
                types.append('int')
            except:
                types.append('str')

    return types


"""Inspect a column/list of data to determine what type of data it contains.

:param values: the list of values to inspect
:param types: (optional) list of types as strings, like ['date', 'int', 'bool']
:param pos: (optional) column position in the dataset, like 0 for the first col
:param key: (optional) column header
:returns a dictional of type probabilities, like {'date': 1, 'int': .75}
"""
def column_types_probabilities(values, types=[], pos=None, key=None):
    all_types = [
        'date', 
        'int', 
        'bool', 
        'float', 
        'category', 
        'id',
        'coord',
        'coord_pair',
        'street',
        'city',
        'region',
        'zipcode',
        'phone',
        'email',
        'url',
        'country',
        'address',
        'text',
        'proportion'
    ]

    types_to_check = types or all_types
    types = {}
    for ttc in types_to_check:
        types[ttc] = column_probability_for_type(values, ttc, pos=pos, key=key)

    return types


"""Get the column_types_probabilities value for each row in a dataset

:param rows: the rows in your dataset
:param headers: a list of column headers (optional)
:param num_rows: max number of rows to use when inspecting types
:returns a list of type probabilities
"""
def rows_types_probabilities(rows, headers=[], num_rows=100):
    types = []
    rows = [row for row in rows if len(row) == len(rows[0])]
    max_rows = num_rows if num_rows < len(rows) else len(rows)

    for i in range(0,len(rows[0])):
        vals = map(lambda x: x[i], rows[:max_rows -1])

        kwargs = { 'pos': i }
        if headers:
            kwargs['key'] = headers[i]

        types.append(column_types_probabilities(vals, **kwargs))

    return types



"""Determine if a column is an address that can be separated into city, state, 
and zipcode

:param values: a list of values
:param num_rows: max number of rows to test
:returns a dict with city, state and zip probabilities
"""
def address_parts_probabilities(values, num_rows=100):
    len_values = len(values)
    ap = AddressParser()
    has = { 'city': 0, 'state': 0, 'zip': 0 }
    probs = { 'city': 0, 'state': 0, 'zip': 0 }
    max_rows = num_rows if num_rows < len_values - 1 else len_values - 1 
        
    if not column_probability_for_type(values[:max_rows], 'address') > .5:
        return probs

    for v in values[:max_rows]:
        if ',' not in v:
            tokens = v.split(' ')
            for i,token in enumerate(tokens):
                if len(token) > 1 and is_a_city(token):
                    tokens[i] = token + ','
                    break

            v = ' '.join(tokens)

        addr = ap.parse_address(v)
        for k in has.keys():
            if getattr(addr, k, None):
                has[k] += 1

    for k in probs.keys():
        probs[k] = float(has[k]) / max_rows

    return probs



def categories_from_list(values):
    non_empty = [r for r in values if r != '']
    threshold = int(.05 * len(non_empty))
    return [x for x, y in collections.Counter(non_empty).items() if y > threshold]



import csv
from .inspectors import (row_simple_types, rows_types_probabilities, 
    categories_from_list)

"""Takes a file object and finds the column headers (if they exist). 

:param fileobj: a file object
:returns found, list of headers
:rtype bool, list
"""
def get_headers(fileobj):
    sniffer = csv.Sniffer()
    sample_contents = fileobj.read(1024)
    dialect = sniffer.sniff(sample_contents)
    fileobj.seek(0)
    reader = csv.reader((line.lstrip() for line in fileobj), dialect=dialect)

    # the csv module sniffer is pretty conversative, so if it thinks we have 
    # a header, let's assume that we do
    if sniffer.has_header(sample_contents):
        return True, reader.next()
    
    # otherwise let's do our own investigation
    rows = list(reader)
    pos_header = rows[0]
    pos_header_types = row_simple_types(pos_header)

    # First make sure each column is a string (headers are always strings)
    if pos_header_types[0] == 'str' and pos_header_types[1:] == pos_header_types[:-1]:
        # now get types for each column
        probs = rows_types_probabilities(rows[1:])
        for i, p in enumerate(probs):
            for k,v in p.iteritems():
                # if the column isn't a string, then its type is different
                # from the first column, so we can assume the first column 
                # is a header
                if v > .5 and k != 'str':
                    # if the column is a category, it might also be 
                    # labeled a string, in which case we should make sure 
                    # the header is not found in the list of categories 
                    # pulled from this dataset
                    if k != 'category':
                        return True, pos_header

                    column_values = map(lambda x: x[i], rows[:1])
                    cats = categories_from_list(column_values)
                    if pos_header[i] not in cats:
                        return True, pos_header 

    # worst case scenario, make up some column names
    return False, ['column_%s' % str(n) for n in range(len(rows[0]))]



from dateutil.parser import parse
from geo_lookup import get_places_by_type
from address import AddressParser
from email.utils import parseaddr
import phonenumbers
import re

def is_a_bool(value, key=None):
    bool_words = ['y','yes','n','no','true','false','t','f','on','off']
    return value.lower().strip() in bool_words


def is_a_date(value, key=None):
    """
    Dateutil recognizes some letter combinations as dates, which is almost 
    always something that isn't really a date (like a US state abbreviation)
    """
    if not any(char.isdigit() for char in value):
        return False

    try:
        pos_date = parse(value)
    except:
        return False

    """ 
    If we can also parse this as a straigtup integer it's probably not a 
    date. Unless of course the word date or time is in the column name, 
    then it might be a timestamp.
    """
    if is_a_int(value, key):
        if not key:
            return False

        keyl = str(key).lower()
        if 'date' in keyl or 'time' in keyl:
            return True
        else:
            return False

    if is_a_float(value, key):
        keyl = str(key).lower()
        if 'date' in keyl or 'time' in keyl:
            return True

        if float(value) < 0:
            return False

        """
        This is iffy. Obviously it's totally possible to have infinitely 
        precise measurements, but we're going to guess that if there are 
        more numbers to the right of the decimal point than the left, we 
        are probably dealing with a coordinate (or something)
        """
        pieces = str(value).split('.')
        if len(pieces[1]) > len(pieces[0]):
            return False

    return True


def is_a_float(value, key=None):
    if "." in str(value):
        try:
            v = float(value)
            return True
        except:
            return False

    return False


def is_a_int(value, key=None):
    if is_a_float(value):
        return False

    if is_a_zip(value, key=key):
        return False

    try:
        int(value)
        return True
    except:
        return False


def is_a_str(value, key=None):
    if not value or str(value).strip() == "":
        return False

    if is_a_date(value) or is_a_float(value) or is_a_int(value):
        return False

    return True


def is_a_text(value, key=None):
    if not is_a_str(value, key=key):
        return False

    return len(str(value).strip()) > 90


"""Geospatial checks. Looks for coordinates, coordinate pairs, address strings, 
geo text (like country, state, city), zip codes, etc etc. Basically anything 
you might want to geocode or treat as a point. """

def is_a_coord(value, key=None, pos=None):
    if key:
        key = str(key).lower().strip()

    if key and key in ['latitude', 'longitude'] and is_a_int(value):
        return True

    if not is_a_int(value, key=key) and not is_a_float(value, key=key):
        return False

    if not abs(float(value)) <= 180:
        return False

    # so we know we have a value that is between -180 and 180
    key_names = ['lat', 'lon', 'lng', 'long', 'coords', 'coordinates']
    if key and any([k in key for k in key_names]):
        return True

    return is_a_float(value, key=key)


def is_a_coord_pair(value, key=None, pos=None):
    delimeters = [',','|','/']
    disallowed = "(){}[]"

    value = str(value).strip()
    possible_matches = [d for d in delimeters if d in value]
    
    # if more than one of these is present or none of them, than this isn't 
    # a pair
    if len(possible_matches) != 1:
        return False

    # Get rid of weird shit people put in coord pair columns
    value = value.translate(None, disallowed)

    delimiter = possible_matches[0]
    possible_cords = value.split(delimiter)
    
    if len(possible_cords) != 2:
        return False

    # All parts have to be floats or ints
    if not all([is_a_float(x) or is_a_int(x) for x in possible_cords]):
        return False

    # max abs lat is 90, max abs lng is 180
    if any([abs(float(x)) > 180 for x in possible_cords]):
        return False

    if all([abs(float(x)) > 90 for x in possible_cords]):
        return False

    """If one is a coord and the other is an int or float, let's use it"""
    if any([is_a_coord(x) for x in possible_cords]):
        return True

    return False
    

def is_a_place(value, place_type, key=None):
    if not is_a_str(value):
        return False

    value = str(value).strip()

    non_addrs = ['|','/','?','!','@','$','%']
    if len([na for na in non_addrs if na in value]) > 0:
        return False

    # If your country's name is longer than 40 characters, you're doing 
    # something wrong.
    if len(value) > 40:
        return False

    if key:
        key = str(key).lower().strip()

    if key and key in [place_type]:
        return True

    if len(get_places_by_type(value, place_type)) > 0:
        return True

    if place_type in ['region'] and len(value) < 4 and \
        len(get_places_by_type(value, place_type + '_iso_code')) > 0:
        return True

    return False


def is_a_city(value, key=None, pos=None):
    return is_a_place(value, 'city', key=key)


def is_a_region(value, key=None, pos=None):
    return is_a_place(value, 'region', key=key)


def is_a_country(value, key=None, pos=None):
    return is_a_place(value, 'country', key=key)


def is_a_zip(value, key=None, pos=None):
    if key:
        key = str(key).lower().strip()

    if key and key in ['zip', 'zipcode', 'postal code']:
        return True

    if str(value).count('-') == 1 and len(str(value)) == 10:
        primary = str(value).split('-')[0]
    else:
        primary = value

    try:
        primary = int(primary)
    except:
        return False

    if len(str(primary)) == 5 and int(primary) > 499:
        return True

    return False


ap = AddressParser()
def address_pieces(value):
    if not is_a_str(value):
        return [], None

    value = str(value).strip()

    if len(value) > 80:
        return [], None

    address = ap.parse_address(value)

    keys = [
        'house_number', 
        'street', 
        'city',
        'zip',
        'state'
    ]

    return [key for key in keys if getattr(address, key, None)], address


"""Check if a string is a house number + street name. The street part of an 
address. Note that we return false if this is a more complete address. """
def is_a_street(value, key=None, pos=None):
    has,address = address_pieces(value)

    if len(has) == 2 and 'house_number' in has and 'street' in has:
        return not is_a_address(value)
    else:
        return False


"""Check to see if this is enough of an address that it could be geocoded. So 
has at least a city + state, or at least street + city"""
def is_a_address(value, key=None, pos=None):
    has,address = address_pieces(value)
    
    if len(has) >= 2:
        if getattr(address, 'city', None):
            return True
        
        pieces = value.split(' ')
        if len(pieces) > 2:
            """Sometimes we get an address like 100 Congress Austin TX, so let's 
            take a stab at breaking up the city/state in a way AddressParser 
            might understand"""
            pieces[len(pieces) - 2] = pieces[len(pieces) - 2] + ','

            has,address = address_pieces(' '.join(pieces))
            if len(has) > 2 and getattr(address, 'city', None):
                return True

    return False


def is_a_phone(value, key=None, pos=None):
    value = str(value).strip()

    if len(value) > 20:
        return False

    """Check for international numbers"""
    if value.startswith('+'):
        try:
            phonenumbers.parse(value)
            return True
        except:
            return False

    """Otherwise let's hope it's a US number"""
    reg = re.compile(".*?(\(?\d{3}\D{0,3}\d{3}\D{0,3}\d{4}).*?", re.S)
    matches = reg.search(value)

    """We're not looking for text fields that contain phone numbers, only fields 
    that are dedicated to phone number"""
    if matches and len(matches.group(1)) == len(value):
        return True


    return False


def is_a_email(value, key=None, pos=None):
    value = str(value).strip()

    possible = parseaddr(value)
    if possible[1] == '':
        return False
    
    e = re.compile(r'[\w\-][\w\-\.]+@[\w\-][\w\-\.]+[a-zA-Z]{1,4}')
    m = e.search(possible[1])
    if not m:
        return False

    if len(m.group(0)) == len(possible[1]):
        return True

    return False


def is_a_url(value, key=None, pos=None):
    # blatantly ripped from Django
    regex = re.compile(
        r'(^https?://)?'  # http:// or https://
        r'(?:(?:[A-Z0-9](?:[A-Z0-9-]{0,61}[A-Z0-9])?\.)+[A-Z]{2,6}\.?|'  # domain...
        r'localhost|'  # localhost...
        r'\d{1,3}\.\d{1,3}\.\d{1,3}\.\d{1,3})' # ...or ip
        r'(?::\d+)?'  # optional port
        r'(?:/?|[/?]\S+)$', re.IGNORECASE)

    value = str(value).strip()
    m = regex.search(value)

    if not m:
        return False

    return len(m.group(0)) == len(value)



import itertools
import math
import collections
from .value_checks import (is_a_date, is_a_int, is_a_bool, is_a_float,
    is_a_coord, is_a_coord_pair, is_a_country, is_a_city, is_a_region,
    is_a_address, is_a_text, is_a_zip, is_a_street, is_a_phone, is_a_url, 
    is_a_email)


"""Guesses likelikhood that a column is of the requested type based on its
values.

:param values: a list of values (probably a column in a csv)
:param for_type: string name of the desired type
:param pos: column position, optionally used for some types
:param key: string name of column header, optionally for some types
:returns probability that this column is of for_type
:rtype float
"""
def column_probability_for_type(values, for_type, pos=None, key=None):
    if for_type == 'category':
        return category_probability(values, key=key, pos=pos)
    elif for_type == 'id':
        return id_probability(values, key=key, pos=pos)
    elif for_type == 'proportion':
        return proportion_probability(values, key=key)

    type_checkers = {
        'date': is_a_date,
        'bool': is_a_bool,
        'int': is_a_int,
        'float': is_a_float,
        'coord': is_a_coord,
        'coord_pair': is_a_coord_pair,
        'street': is_a_street,
        'city': is_a_city,
        'region': is_a_region,
        'zipcode': is_a_zip,
        'phone': is_a_phone,
        'url': is_a_url,
        'email': is_a_email,
        'country': is_a_country,
        'address': is_a_address,
        'text': is_a_text
    }

    is_type = 0
    for value in values:
        if type_checkers[for_type](value):
            is_type += 1


    non_empty = [v for v in values if v != '']
    divisor = len(non_empty) if len(non_empty) > 0 else 1

    return float(is_type) / divisor


"""The likelihood that a given list of values is a unique id

:param values: a list of values to check
:param key: the column header (optional)
:param pos: the column position (optiona - eg 0 for the first column)
"""
def id_probability(values, key=None, pos=None):
    prob = 0

    if len(values) == len(set(values)):
        prob += 0.25
    else:
        # if val in all rows isn't unique, then it's not an id
        return 0

    if key and str(key).lower() == 'id':
        prob += 0.5

    if pos and pos == 0:
        prob += .25

    return prob

"""The likelihood that a given list of values is a proportion

:param values: a list of values to check
:param key: the column header (optional)
:param pos: the column position (optiona - eg 0 for the first column)
"""
def proportion_probability(values, key=None, pos=None):
    if column_probability_for_type(values, 'int') == 1:
        try:
            values = [int(v) for v in values]
        except:
            return 0
    elif column_probability_for_type(values, 'float') == 1:
        try:
            values = [float(v) for v in values]
        except:
            return 0
    else:
        return 0

    if sum(values) == 1 or sum(values) == 100:
        return 1

    return 0


"""The likelihood that a given list of values is a category. A category column
is assumed to have the following properties:

- most of its rows are not empty
- it's not populated with dates
- any given category label should be assigned to at least 2% of the rows
- 80% of the rows should be assigned to at least one category label, or...
- at least 10% of the rows must be assigned to one category label

It's not perfect, but it usually finds columns that you'd want to group as
categories when further analyzing a dataset.

:param values: a list of values to check
:param key: the column header (optional)
:param pos: the column position (optiona - eg 0 for the first column)
"""
def category_probability(values, key=None, pos=None):
    total_rows = len(values)
    non_empty = [r.strip() for r in values if r.strip() != '']

    # If this is a mostly empty rows, we don't want to use it
    if len(non_empty) < len(values) / 2:
        return 0

    # If this is a date field, we don't treat it as a category
    if column_probability_for_type(values, 'date', pos=pos, key=key) > .5:
        return 0

    # All category fields are 1 to many. If we detect a delimiter, split the
    # text field into a list. Otherwise treat the category as a list of 1
    delimiter = detect_delimiter(values)
    if delimiter:
        row_cats = [i.split(delimiter) for i in non_empty]
        all_cats = list(itertools.chain.from_iterable(row_cats))
    else:
        row_cats = [[x] for x in non_empty]
        all_cats = non_empty


    # at least 2% of the rows in the set must have any given value for it to
    # be considered a possible category
    threshold = int(.02 * len(non_empty))

    # number of categories
    matches = []
    failed_matches = []

    # This prodcues a list of two-item lists, like [String, Integer], where
    # String is the category label and Integer is the number of times that
    # value appears in the list
    for x,y in collections.Counter(all_cats).items():
        # y > 1 for small datasets (like 25 rows) where 1 row is more than
        # 2% of overall rows.
        if y > 1 and y > threshold:
            matches.append([x,y])
        else:
            failed_matches.append([x,y])

    if len(matches) == 0:
        return 0

    # Determine the number of rows that have been assigned to at least one
    # the values we determined to be a category using the 2% test above.
    total_categorized = 0
    total_uncategorized = 0

    all_matches = [x[0] for x in matches]
    for row_cat in row_cats:
        if len([x for x in row_cat if x in all_matches]) > 0:
            total_categorized += 1
        else:
            total_uncategorized += 1


    # If this is a bunch of numbers, it's probably not a category
    if all([is_a_int(v) or is_a_float(v) for v in values]) and \
            len(all_matches) > math.sqrt(total_rows) / 2:

        return 0 


    # If more than half the dataset isn't categorized based on this set of
    # labels, then it's probably not a category

    if float(total_uncategorized) / float(total_rows) >= .5:
        return 0

    # more than 20% of the rows aren't categorized, and none of the categories
    # are more than 10% of the dataset
    elif float(total_uncategorized) / float(total_rows) >= .2 and \
            all([float(x[1]) / float(len(non_empty)) < .1 for x in matches]):

        return 0
    else:
        tu = total_uncategorized if total_uncategorized > 0 else 1

        prob = float(total_categorized) / float(tu)
        uncategorized_ratio = float(tu) / float(total_categorized)

        print uncategorized_ratio
        if uncategorized_ratio > .5:
            return 0

        if prob >= 1:
            return 1

        return prob


"""Finds the delimeter used to separate values in a string. This is for
category columns in a dataset like a,b,c or a/b/c

:param values: list of values (probably a column from a csv)
:returns the delimiter or None
:rtype string or None
"""
def detect_delimiter(values):
    non_empty = [r for r in values if r != '']

    # If this is a mostly empty rows, we don't want to use it
    if len(non_empty) < len(values) / 2:
        return None

    delimiters = [',','|','/']

    # 50 is a wild guess, should be refined
    short_enough = [v for v in non_empty if len(v) < 50]
    if len(short_enough) < len(non_empty) / 2:
        return None

    # number of records with each type of delimiter
    delimeter_count = [0] * len(delimiters)
    for i,d in enumerate(delimiters):
        delimeter_count[i] = len([v for v in non_empty if d in v])

    # if more than half the records contain one type of delimeter, it wins
    possible_delimeters = []
    for i,dc in enumerate(delimeter_count):
        if dc > len(non_empty) / 2:
            possible_delimeters.append(delimiters[i])

    if len(possible_delimeters) != 1:
        return None

    delimiter = possible_delimeters[0]

    def average(s): return sum(s) * 1.0 / len(s)

    # get average word length and standard deviation
    categories = list(itertools.chain.from_iterable([i.split(delimiter) for i in non_empty]))
    word_lengths = [len(w) for w in categories]
    avg_word_length = average(word_lengths)
    variance = map(lambda x: (x - avg_word_length)**2, word_lengths)
    standard_deviation = math.sqrt(average(variance))

    # this was probably a text field that used commas (or whatever) as part
    # of sentences.
    if standard_deviation > 6:
        return None

    return delimiter


from setuptools import setup
import os

try:
   import pypandoc
   long_description = pypandoc.convert('README.md', 'rst')
except (IOError, ImportError):
   long_description = open('README.md').read()

setup(name='penny',
      version='0.3.9',
      description='Inspect your data. Find the truth!',
      long_description=long_description,
      url='https://github.com/gati/penny',
      download_url ='https://github.com/gati/penny/tarball/0.3.9',
      author='Jonathon Morgan',
      author_email='jonathon@goodattheinternet.com',
      license='MIT',
      packages=['penny'],
      test_suite='tests',
      install_requires=['python-dateutil','address','phonenumbers'],
      package_data = {
            'penny': ['data/*.csv'],
      },
      zip_safe=False)